@@ -28,7 +28,8 @@ def dict_from_gamut(input_dir):
     if ext == FILE_EXT:
         return np.load(input_dir, allow_pickle=True).item()
     else:
-        raise ValueError('Wrong file extension. Provide a directory for a {} file'.format(FILE_EXT))
+        raise ValueError(
+            'Wrong file extension. Provide a directory for a {} file'.format(FILE_EXT))
 
 def write_audio(output_dir, ndarray, sr=44100, bit_depth=24):
     """Writes a `ndarray` as audio. This function is a simple wrapper of `sf.write()`."""
@@ -67,18 +68,19 @@ def get_features(input_dir, max_duration=None, n_mfcc=13, hop_length=512, frame_
                                                                 hop_length=hop_length)[0][:-2]/440)*12+69
     centroid_frames[centroid_frames < 0] = 0
     n_frames = len(rms_frames)
-    sample_indxs = np.arange(start=0, stop=n_frames * hop_length, step=hop_length)
+    sample_indxs = np.arange(start=0, stop=n_frames *
+                             hop_length, step=hop_length)
     metadata = np.array([sample_indxs, rms_frames, centroid_frames]).T
 
     return mfcc_frames, metadata, sr
 
 def build_corpus(input_dir, max_duration=None, n_mfcc=13, hop_length=512, frame_length=1024, kd=None):
     '''Takes a folder directory, or an audio file directory, or a list of directories to audio files, and returns a `dict` object. The output can be saved as a `.gamut` file with the `dict_to_gamut()` function, for later use in `get_audio_recipe()`.'''
-   
-    soundfiles = list() 
-   
+
+    soundfiles = list()
+
     # check if input_dir is folder or list of s
-    if type(input_dir) is str: 
+    if type(input_dir) is str:
         input_dir = realpath(input_dir)
         if isdir(input_dir):
             read_mode = 'folder'
@@ -121,16 +123,19 @@ def build_corpus(input_dir, max_duration=None, n_mfcc=13, hop_length=512, frame_
 
     # if input is a folder directory, collect all files from folder and append to soundfiles list
     if read_mode == 'folder':
-        [[soundfiles.append(join(root, f)) for f in files] for root, _, files in walk(input_dir)]
+        [[soundfiles.append(join(root, f)) for f in files]
+         for root, _, files in walk(input_dir)]
 
     # iterate through paths in soundfiles, select audio files, and extract features
     for sf in soundfiles:
         sf = realpath(sf)
         if splitext(sf)[1] in AUDIO_FORMATS:
-            mfcc_stream, metadata, sr = get_features(sf, max_duration=max_duration, n_mfcc=n_mfcc, hop_length=hop_length, frame_length=frame_length)
+            mfcc_stream, metadata, sr = get_features(
+                sf, max_duration=max_duration, n_mfcc=n_mfcc, hop_length=hop_length, frame_length=frame_length)
             for mf, md in zip(mfcc_stream, metadata):
                 mfcc_frames.append(mf)
-                dictionary['data_samples'].append(np.array(np.concatenate([[file_id],  md])))
+                dictionary['data_samples'].append(
+                    np.array(np.concatenate([[file_id],  md])))
             dictionary['corpus_info']['files'].append([sr, sf])
             counter.write(str(file_id + 1))
             file_id += 1
@@ -145,20 +150,63 @@ def build_corpus(input_dir, max_duration=None, n_mfcc=13, hop_length=512, frame_
     print('        DONE\n')
     return dictionary
 
+def find_nodes(data, depth=0, kd=None, medians=None, nodes=list(), count=1):
+	N = len(data)
+	# initialize nodes
+	if len(nodes) == 0:
+		nodes = [[]]*kd	
+    # pass global median to branch if branch is empty
+	if N <= 0:
+		if depth < kd:
+			nodes[depth] = nodes[depth]+[[medians[depth], count]]		
+		return nodes
+    # split branches recursively when depth < kd
+	if depth < kd:
+		sorted_data = data[data[:, depth].argsort()]
+		split = floor(int(N/2))
+		node = sorted_data[split][depth]
+		if node is not None:
+			nodes[depth] = nodes[depth]+[[node, count]]
+		find_nodes(sorted_data[:split], depth+1, kd=kd, medians=medians, nodes=nodes, count=count*2)
+		find_nodes(sorted_data[split + 1:], depth+1, kd=kd, medians=medians, nodes=nodes, count=count*2+1)
+		return nodes
+	else:
+		return nodes # stop recursion
+
+def sort_nodes(tree_nodes, medians=None):
+	return [[(n[j][0] if n[j][1] == (2**i)+j else medians[i]) if j < len(n) else medians[i] for j in range(2**i)] for i, n in enumerate(tree_nodes)]
+
+def get_branch_id(vector, nodes):
+    idx = 0
+    for i, v in enumerate(vector):
+        node = nodes[i][idx]
+        idx = (idx << 1) | 1 if v > node else idx << 1
+    return idx
+
+def fit_data(data, data_min=list(), data_max=list()):
+    if len(data_min) * len(data_max) == 0:
+        data_min = np.amin(data, axis=0)
+        data_max = np.amax(data, axis=0)
+    return (data-data_min)/(data_max-data_min), data_min, data_max
+
 def build_data_tree(data, kd=2):
-    '''Creates a binary tree data structure.'''
+    '''Creates a KDTree data structure.'''
     print()
-    data = np.array(data)
-    bar = IncrementalBar('        Classifying data frames: ', max=len(data), suffix='%(index)d/%(max)d frames')
-    nodes = np.median(data[:,:kd], axis=0)
-    tree_size = 2**len(nodes)
+    data, data_min, data_max = fit_data(np.array(data))
+    bar = IncrementalBar('        Classifying data frames: ', max=len(
+        data), suffix='%(index)d/%(max)d frames')
+    medians = np.median(data, axis=0)
+    nodes = sort_nodes(find_nodes(data, kd=kd, medians=medians), medians=medians)
+    tree_size = 2**len(data[0][:kd])
     tree = {
         'nodes': nodes,
         'dims': kd,
+        'data_min': data_min,
+        'data_max': data_max,
         'position_branches': dict(),
         'data_branches': dict()
     }
-    # initialize branches
+    # populate branches
     for i in range(tree_size):
         tree['position_branches'][str(i)] = list()
         tree['data_branches'][str(i)] = list()
@@ -176,10 +224,6 @@ def build_data_tree(data, kd=2):
     print("        Total number of data clusters: {}".format(tree_size))
     return tree
 
-def get_branch_id(vector, nodes):
-    size = len(vector)-1
-    return sum([(2<<(size-i))>>1 if v > n else 0 for i, (v, n) in enumerate(zip(vector, nodes))])
-
 def neighborhood_index(item, tree):
     kd = tree['dims']
     tree_size = 2**kd
@@ -195,6 +239,7 @@ def wedgesum(a, b):
     b = (b * -1) if (b % 2 == 0) else (b + 1)
     return a + floor(b/2)
 
+
 def wrap(a, min, max):
     return ((a-min) % max-min) + min
 
@@ -203,18 +248,20 @@ def nearest_neighbors(item, data, kn=8):
     if datasize < kn:
         kn = datasize
     nn = NearestNeighbors(n_neighbors=kn, algorithm='brute').fit(data)
-    positions = nn.kneighbors(item, n_neighbors=kn)[1][0]
-    return positions
+    dx, positions = nn.kneighbors(item, n_neighbors=kn)
+    return positions[0], dx[0]
 
 def get_audio_recipe(target_dir, corpus_dict, max_duration=None, hop_length=512, frame_length=1024, kn=8):
     '''Takes an audio sample directory/path (i.e. the _target_) and a `dict` object (i.e. the _corpus_), and returns another `dict` object containing the instructions to rebuild the _target_ using grains from the _corpus_. The output can be saved as a `.gamut` file with the `dict_to_gamut()` function, for later use in `cook_recipe()`.'''
 
-    print('\nMaking recipe for {}\n        ...loading corpus...\n        ...analyzing target...'.format(basename(target_dir)))
+    print('\nMaking recipe for {}\n        ...loading corpus...\n        ...analyzing target...'.format(
+        basename(target_dir)))
     target_mfcc, target_extras, sr = get_features(target_dir,
                                                   max_duration=max_duration,
                                                   n_mfcc=corpus_dict['corpus_info']['n_mfcc'],
                                                   hop_length=hop_length,
                                                   frame_length=frame_length)
+    target_mfcc = fit_data(target_mfcc, data_min=corpus_dict['data_tree']['data_min'], data_max=corpus_dict['data_tree']['data_max'])[0]
     n_samples = len(target_extras)
     dictionary = {
         'target_info': {
@@ -236,24 +283,27 @@ def get_audio_recipe(target_dir, corpus_dict, max_duration=None, hop_length=512,
 
     # include target data samples for cooking mix parameter
     corpus_dict['corpus_info']['files'].append([sr, target_dir])
-    dictionary['target_info']['data_samples'] = target_extras[:,[0, 0]].astype('int32')
+    dictionary['target_info']['data_samples'] = target_extras[:,
+                                                              [0, 0]].astype('int32')
     dictionary['target_info']['data_samples'][:, 0] = len(
         corpus_dict['corpus_info']['files']) - 1
 
-    bar = IncrementalBar('        Matching audio frames: ', max=len(
-        target_mfcc), suffix='%(index)d/%(max)d frames')
-
+    error_sum = 0
+    bar = IncrementalBar('        Matching audio frames: ', max=len(target_mfcc), suffix='%(index)d/%(max)d frames')
     for tm, tex in zip(target_mfcc, target_extras):
         branch_id = str(neighborhood_index(tm, corpus_dict['data_tree']))
-        mfcc_idxs = nearest_neighbors([tm], corpus_dict['data_tree']['data_branches'][branch_id], kn=kn)
+        mfcc_idxs, dx = nearest_neighbors([tm], corpus_dict['data_tree']['data_branches'][branch_id], kn=kn)
+        error_sum += dx[0]
         knn_positions = corpus_dict['data_tree']['position_branches'][branch_id][mfcc_idxs]
         metadata = corpus_dict['data_samples'][knn_positions]
-        sorted_positions = nearest_neighbors([tex[1:]], metadata[:, [2, 3]], kn=kn)
+        sorted_positions = nearest_neighbors([tex[1:]], metadata[:, [2, 3]], kn=kn)[0]
         mfcc_options = metadata[sorted_positions]
-        dictionary['data_samples'].append(mfcc_options[:, [0, 1]].astype('int32'))
+        dictionary['data_samples'].append(
+            mfcc_options[:, [0, 1]].astype('int32'))
         bar.next()
     bar.finish()
-    print('        DONE\n')
+    error_sum = int((error_sum/n_samples)*1000)/1000
+    print('        Delta average: {}\n        DONE.\n'.format(error_sum))
     return dictionary
 
 def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, target_mix=0, pan_spread=0.5, kn=8, n_chans=2, envelope='hann', sr=None, frame_length_res=512):
@@ -272,11 +322,14 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
     snd_idxs = list()
     [[snd_idxs.append(y[0]) for y in x] for x in recipe_dict['data_samples']]
     snd_idxs = np.unique(snd_idxs).astype(int)
-    snd_counter = IncrementalBar('        Loading corpus sounds: ', max=len(snd_idxs) + 1, suffix='%(index)d/%(max)d files')
+    snd_counter = IncrementalBar('        Loading corpus sounds: ', max=len(
+        snd_idxs) + 1, suffix='%(index)d/%(max)d files')
     for i in snd_idxs:
-        sounds[i] = np.repeat(np.array([librosa.load(recipe_dict['corpus_info']['files'][i][1], duration=recipe_dict['corpus_info']['max_duration'], sr=sr)[0]]).T, n_chans, axis=1)
+        sounds[i] = np.repeat(np.array([librosa.load(recipe_dict['corpus_info']['files'][i][1],
+                              duration=recipe_dict['corpus_info']['max_duration'], sr=sr)[0]]).T, n_chans, axis=1)
         snd_counter.next()
-    sounds[-1] = np.repeat(np.array([librosa.load(recipe_dict['corpus_info']['files'][-1][1], duration=None, sr=sr)[0]]).T, n_chans, axis=1)
+    sounds[-1] = np.repeat(np.array([librosa.load(recipe_dict['corpus_info']
+                           ['files'][-1][1], duration=None, sr=sr)[0]]).T, n_chans, axis=1)
     snd_counter.next()
     snd_counter.finish()
 
@@ -296,12 +349,15 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
     # populate frame length table
     grain_dur_type = type(grain_dur)
     if grain_dur_type is list:
-        frame_length_table = np.round(array_resampling(np.array(grain_dur) * sr, n_segments)/frame_length_res) * frame_length_res
+        frame_length_table = np.round(array_resampling(
+            np.array(grain_dur) * sr, n_segments)/frame_length_res) * frame_length_res
     if grain_dur_type is float or grain_dur_type is int:
         frame_length_table = np.empty(n_segments)
-        frame_length_table.fill(round((grain_dur * sr)/frame_length_res)*frame_length_res)
+        frame_length_table.fill(
+            round((grain_dur * sr)/frame_length_res)*frame_length_res)
     frame_length_table = frame_length_table.astype('int64')
-    frame_lengths = np.arange(frame_length_res, np.amax(frame_length_table)+frame_length_res, frame_length_res, dtype='int64')
+    frame_lengths = np.arange(frame_length_res, np.amax(
+        frame_length_table)+frame_length_res, frame_length_res, dtype='int64')
 
     # populate sample index table
     stretch_type = type(stretch_factor)
@@ -309,27 +365,33 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
         samp_onset_table = np.empty(n_segments)
         samp_onset_table.fill(hop_length*stretch_factor)
     if stretch_type is list:
-        samp_onset_table = array_resampling(stretch_factor, n_segments)*hop_length
-    samp_onset_table = np.concatenate([[0], np.round(samp_onset_table)], ).astype('int64').cumsum()[:-1]
+        samp_onset_table = array_resampling(
+            stretch_factor, n_segments)*hop_length
+    samp_onset_table = np.concatenate(
+        [[0], np.round(samp_onset_table)], ).astype('int64').cumsum()[:-1]
 
     # apply onset variation table
     onset_var_type = type(onset_var)
     if onset_var_type is float or onset_var_type is int:
         if onset_var > 0:
             jitter = int(max(1, abs((onset_var*sr)/2)))
-            onset_var_table = np.random.randint(low=jitter*-1, high=jitter, size=n_segments)
+            onset_var_table = np.random.randint(
+                low=jitter*-1, high=jitter, size=n_segments)
             samp_onset_table = samp_onset_table + onset_var_table
     if onset_var_type is list:
-        onset_var_table = array_resampling(np.array(onset_var)*sr, n_segments) * np.random.rand(n_segments)
+        onset_var_table = array_resampling(
+            np.array(onset_var)*sr, n_segments) * np.random.rand(n_segments)
         samp_onset_table = samp_onset_table + onset_var_table.astype('int64')
     samp_onset_table[samp_onset_table < 0] = 0
 
     # compute window with default size
     env_type = type(envelope)
     if env_type == str:
-        windows = [np.repeat(np.array([get_window(envelope, Nx=wl)]).T, n_chans, axis=1) for wl in frame_lengths]
+        windows = [np.repeat(np.array(
+            [get_window(envelope, Nx=wl)]).T, n_chans, axis=1) for wl in frame_lengths]
     if env_type == np.ndarray or env_type == list:
-        windows = [np.repeat(np.array([array_resampling(envelope, N=wl)]).T, n_chans, axis=1) for wl in frame_lengths]
+        windows = [np.repeat(np.array(
+            [array_resampling(envelope, N=wl)]).T, n_chans, axis=1) for wl in frame_lengths]
 
     # compute panning table
     pan_type = type(pan_spread)
@@ -340,7 +402,8 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
         pan_table = np.power(pan_table, pan_spread)
     if pan_type is list:
         pan_spread = np.array(pan_spread)
-        pan_table = np.power(pan_table, np.repeat(np.array([array_resampling(pan_spread, n_segments)]).T, n_chans, axis=1))
+        pan_table = np.power(pan_table, np.repeat(
+            np.array([array_resampling(pan_spread, n_segments)]).T, n_chans, axis=1))
     row_sums = pan_table.sum(axis=1)
     pan_table = pan_table / row_sums[:, np.newaxis]
 
@@ -349,10 +412,12 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
     weigths = np.arange(kn, 0, -1)
 
     # make buffer array
-    buffer = np.empty(shape=(int(np.amax(samp_onset_table) + np.amax(frame_length_table)), n_chans))
+    buffer = np.empty(
+        shape=(int(np.amax(samp_onset_table) + np.amax(frame_length_table)), n_chans))
     buffer.fill(0)
 
-    grain_counter = IncrementalBar('        Concatenating grains:  ', max=len(data_samples), suffix='%(index)d/%(max)d grains')
+    grain_counter = IncrementalBar('        Concatenating grains:  ', max=len(
+        data_samples), suffix='%(index)d/%(max)d grains')
 
     for n, (ds, so, fl, p, tm) in enumerate(zip(data_samples, samp_onset_table, frame_length_table, pan_table, target_mix_table)):
         if random() > tm:
@@ -381,4 +446,4 @@ def cook_recipe(recipe_dict, grain_dur=0.1, stretch_factor=1, onset_var=0, targe
     return (buffer / np.amax(np.abs(buffer))) * sqrt(0.5)
 
 if __name__ == '__main__':
-    print()
\ No newline at end of file
+    print()
